%learn incoherent dictionary.
%
% Y is the data as dxn matrix, where d is the dimensionality of the data and n the number of data pts.
% numbasis: the number of atoms in the dictioanry
% beta: the  incoherence value
% lambda: the L2 error for the lasso
% D_learn: any initial dicitonary, [] if nothing is available
% max_active_components : number of active atoms allowed in the sparse code (param.L in mexLasso).
% for bugs contact anoop.cherian@inria.fr
%
function D_learn = incoherent_dl(Y, numbasis, beta, lambda, D_learn, max_active_components)
dim = size(Y,1);
mu_min = sqrt((numbasis-dim)/(dim*(numbasis-1)));
fprintf('mu=%f min coherence possible=%f\n', beta, mu_min);
if beta < mu_min
   error('expeceted coherence is less than the minimum coherence possible with this dictionary configuration...\n');
end

if isempty(D_learn)
    fprintf('D_learn is null\n');
    perm = randperm(size(Y,2));
    D_learn = Y(:,perm(1:numbasis));
    D_learn = bsxfun(@rdivide, D_learn, sqrt(diag(D_learn'*D_learn))');
end

obj_old = 1e10; D = 0;
[d,n] = size(D_learn);
L = [eye(n-1), -eye(n-1)];

for j=1:100
    X = sparse_code(D_learn, Y, lambda, max_active_components);
    num_iter = 10 + ceil(j/10)*5;
    D_learn = inco_dict_learn(Y, X, D_learn, L, beta, num_iter);
    obj = compute_obj(D_learn, Y, X, lambda, beta);
    fprintf('iter-%d) ||D_learn-D||_2 = %f, obj=%f, mu(D) = %f\n', j,...
                    norm(D_learn-D,'fro'), obj, compute_mu(D_learn));
    if ~exist('./idltemp/', 'dir'), mkdir('./idltemp/'); end
    save(['./idltemp/inco_D_' num2str(size(D_learn,2)) '_mu' num2str(beta) 'l_' num2str(lambda) '.mat'], 'D_learn');
    if abs(obj_old - obj) < 0.01 || norm(D_learn - D, 'fro') < 1e-2
        break;
    end
    obj_old = obj;
end
fprintf('finished\n');
end

% use your favorite sparse coding algorithm
function X = sparse_code(D, Y, lambda, max_active_components)
%fprintf('spcode\n');
param.mode=1; param.L = max_active_components; param.lambda=lambda; param.lambda2=0; param.numThreads=-1;
X = mexLasso(Y, D, param);
%fprintf('spcode end\n');
end

% incoherent dl.
function D_learn = inco_dict_learn(Y, X, D, L, beta, num_iter)
[d,n] = size(D);  %beta = 1/tau + beta;
I = 1:n; D_old = D; dim = d; Z = zeros(2*(n-1),1);%
a = zeros(n,1);
for k=1:n
    a(k) = 1/(sum(X(k,:).^2)+eps);
end
Lambda = zeros(2*(n-1), n); % a cache for the dual atoms for speed up. 
S = Y - D*X; % save the difference for speed
one = ones(dim,1);
for it=1:num_iter
    tic;
    for k=1:n        
        if mod(k,32)==0, 
 	   fprintf('.');
        end
        D_til = D; D_til(:,k) = []; atom = D(:,k);
        nzidx = find(abs(X(k,:))>1e-6);
        XXk = X(k,nzidx);
        Y_til = S(:,nzidx) + atom*XXk;
        ones_Xk = one*XXk;
        theta = sum(Y_til.*ones_Xk, 2);        
        
        %Lambda(:,k) = mxFastBBNNLS(a(k), [D_til, -D_til], theta, beta, Lambda(:,k), 75);
        Lambda(:,k) = fastbbnnls(a(k), [D_til, -D_til], theta, beta, Lambda(:,k), 200);

        dlambda = Lambda(1:n-1,k) - Lambda(n:2*(n-1),k);
        x = a(k)*(theta + D_til*dlambda);
        x = x/(norm(x)+eps);
        S(:,nzidx) = Y_til - x*XXk;
        
        D(:,k) = x;
    end
      
    fprintf('%d) DL convergence: obj=%f took %f seconds\n', it, norm(D-D_old,'fro'), toc);
    if norm(D - D_old, 'fro') < 1e-2
        break;
    end
    D_old = D;
end
D_learn = D;
end

% objective function
function obj = compute_obj(D_learn, Y, X, lambda, beta)
n = size(D_learn,2); coh_sum = 0;
for i=1:n
    coh_sum = coh_sum + norm(D_learn'*D_learn(:,i),inf);
end
obj = 0.5*norm(Y-D_learn*X,'fro')^2 + lambda*sum(sum(abs(X))) + beta*coh_sum;
end

% coherence
function mu = compute_mu(D_learn)
n = size(D_learn,2);
DD = D_learn'*D_learn-eye(n);
mu = max(abs(DD(:)));
%mu = sum(sum(abs(DD)))/2/(n*(n-1))/2;
end
